use crate::{
    intern::{Intern, Interned},
    source_location::SourceLocation,
};
use num_bigint::BigInt;
use std::{
    convert::Infallible,
    fmt,
    hash::Hash,
    marker::PhantomData,
    ops::{Deref, DerefMut, Index, IndexMut},
};

pub(crate) type SmallUInt = u64;
pub(crate) type SmallSInt = i64;

macro_rules! impl_insns {
    (
        #[insn = $Insn:ident, next_macro = $next_macro:ident, branch_macro = $branch_macro:ident]
        $vis:vis fn $State:ident::$run:ident(&mut $self:ident) -> $run_ret_ty:ty {
            #[state]
            let mut $state:ident = $state_init:expr;
            setup! {
                $($setup:tt)*
            }
            main_loop!();
            cleanup! {
                $($cleanup:tt)*
            }
        }
        $(
            $(#[$insn_meta:meta])*
            $insn_name:ident $({
                $(
                    $(#[$field_meta:meta])*
                    $field_name:ident: $field_ty:ty,
                )*
            })? => $block:block
        )+
    ) => {
        #[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
        $vis enum $Insn {
            $(
                $(#[$insn_meta])*
                $insn_name $({
                    $(
                        $(#[$field_meta])*
                        $field_name: $field_ty,
                    )*
                })?,
            )+
        }

        impl $State {
            $vis fn $run(&mut $self) -> $run_ret_ty {
                let mut $state = $state_init;
                $($setup)*
                let mut insn = $state.insns[$state.pc];
                let retval = 'main_loop: loop {
                    macro_rules! $next_macro {
                        () => {
                            $state.pc += 1;
                            insn = $state.insns[$state.pc];
                            continue 'main_loop;
                        };
                    }
                    macro_rules! $branch_macro {
                        ($next_pc:expr) => {
                            $state.pc = $next_pc;
                            insn = $state.insns[$state.pc];
                            continue 'main_loop;
                        };
                    }
                    let _: Infallible = match insn {
                        $(
                            $Insn::$insn_name $({
                                $(
                                    $field_name,
                                )*
                            })? => {
                                $block
                            }
                        )+
                    };
                };
                $($cleanup)*
                retval
            }
        }
    };
}

pub(crate) trait InsnsBuildingKind: Copy + Eq + fmt::Debug + Hash + Default {
    type Vec<T: fmt::Debug + Clone + Eq + Hash + Send + Sync + 'static>: fmt::Debug
        + Clone
        + Eq
        + Hash
        + Send
        + Sync
        + 'static
        + Default
        + Deref<Target = [T]>
        + FromIterator<T>;
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
pub(crate) struct InsnsBuildingDone;

impl InsnsBuildingKind for InsnsBuildingDone {
    type Vec<T: fmt::Debug + Clone + Eq + Hash + Send + Sync + 'static> = Interned<[T]>;
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
pub(crate) struct InsnsBuilding;

impl InsnsBuildingKind for InsnsBuilding {
    type Vec<T: fmt::Debug + Clone + Eq + Hash + Send + Sync + 'static> = Vec<T>;
}

#[derive(Clone, PartialEq, Eq, Hash)]
pub(crate) struct Insns<BK: InsnsBuildingKind> {
    insns: BK::Vec<Insn>,
    insn_source_locations: BK::Vec<SourceLocation>,
    state_layout: StateLayout<BK>,
}

struct InsnsDebug<'a> {
    insns: &'a [Insn],
    insn_source_locations: &'a [SourceLocation],
}

impl<'a> fmt::Debug for InsnsDebug<'a> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let mut debug_list = f.debug_list();
        let mut last_source_location = None;
        for (insn, source_location) in self.insns.iter().zip(self.insn_source_locations) {
            if Some(source_location) != last_source_location {
                debug_list.entry(&format_args!("// at: {source_location}\n{insn:?}"));
            } else {
                debug_list.entry(insn);
            }
            last_source_location = Some(source_location);
        }
        debug_list.finish()
    }
}

impl<BK: InsnsBuildingKind> fmt::Debug for Insns<BK> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            insns,
            insn_source_locations,
            state_layout,
        } = self;
        f.debug_struct("Insns")
            .field("state_layout", state_layout)
            .field(
                "insns",
                &InsnsDebug {
                    insns,
                    insn_source_locations,
                },
            )
            .finish_non_exhaustive()
    }
}

pub(crate) trait StatePartKind:
    Send + Sync + Ord + Hash + fmt::Debug + 'static + Copy + Default
{
    const NAME: &'static str;
    type DebugData: Send + Sync + Eq + Hash + fmt::Debug + 'static + Copy;
    type State: fmt::Debug + 'static + Clone;
    type BorrowedState<'a>: 'a;
    fn new_state(len: StatePartLen<Self>) -> Self::State;
    fn borrow_state<'a>(state: &'a mut Self::State) -> Self::BorrowedState<'a>;
}

macro_rules! make_state_part_kinds {
    (
        $(
            #[state, field = $state_field:ident]
            impl $StateStatePartKind:ident for $StateKind:ident $state_impl_body:tt
        )+
        $(
            #[type, field = $type_field:ident]
            impl $TypeStatePartKind:ident for $TypeKind:ident $type_impl_body:tt
        )+
    ) => {
        $(
            #[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy, Default)]
            pub(crate) struct $StateKind;

            impl $StateStatePartKind for $StateKind $state_impl_body
        )+

        $(
            #[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Debug, Clone, Copy, Default)]
            pub(crate) struct $TypeKind;

            impl $TypeStatePartKind for $TypeKind $type_impl_body
        )+

        #[derive(Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct StateLayout<BK: InsnsBuildingKind> {
            pub(crate) ty: TypeLayout<BK>,
            $(pub(crate) $state_field: StatePartLayout<$StateKind, BK>,)+
        }

        impl Copy for StateLayout<InsnsBuildingDone> {}

        impl<BK: InsnsBuildingKind> StateLayout<BK> {
            pub(crate) fn len(self) -> StateLen {
                StateLen {
                    ty: self.ty.len(),
                    $($state_field: self.$state_field.len(),)+
                }
            }
            pub(crate) fn is_empty(self) -> bool {
                self.ty.is_empty() $(&& self.$state_field.is_empty())+
            }
            pub(crate) fn empty() -> Self {
                Self {
                    ty: TypeLayout::empty(),
                    $($state_field: StatePartLayout::empty(),)+
                }
            }
        }

        impl StateLayout<InsnsBuilding> {
            pub(crate) fn next_index(&self) -> StateIndex {
                StateIndex {
                    ty: self.ty.next_index(),
                    $($state_field: self.$state_field.next_index(),)+
                }
            }
            pub(crate) fn allocate<BK: InsnsBuildingKind>(
                &mut self,
                layout: &StateLayout<BK>,
            ) -> StateIndexRange {
                StateIndexRange {
                    ty: self.ty.allocate(&layout.ty),
                    $($state_field: self.$state_field.allocate(&layout.$state_field),)+
                }
            }
        }

        impl From<StateLayout<InsnsBuilding>> for StateLayout<InsnsBuildingDone> {
            fn from(v: StateLayout<InsnsBuilding>) -> Self {
                Self {
                    ty: v.ty.into(),
                    $($state_field: v.$state_field.into(),)+
                }
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct StateIndexRange {
            pub(crate) ty: TypeIndexRange,
            $(pub(crate) $state_field: StatePartIndexRange<$StateKind>,)+
        }

        impl StateIndexRange {
            pub(crate) fn start(self) -> StateIndex {
                StateIndex {
                    ty: self.ty.start(),
                    $($state_field: self.$state_field.start(),)+
                }
            }
            pub(crate) fn len(self) -> StateLen {
                StateLen {
                    ty: self.ty.len(),
                    $($state_field: self.$state_field.len(),)+
                }
            }
            pub(crate) fn is_empty(self) -> bool {
                self.ty.is_empty() $(&& self.$state_field.is_empty())+
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct StateLen {
            pub(crate) ty: TypeLen,
            $(pub(crate) $state_field: StatePartLen<$StateKind>,)+
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct StateIndex {
            pub(crate) ty: TypeIndex,
            $(pub(crate) $state_field: StatePartIndex<$StateKind>,)+
        }

        #[derive(Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct TypeLayout<BK: InsnsBuildingKind> {
            $(pub(crate) $type_field: StatePartLayout<$TypeKind, BK>,)+
        }

        impl Copy for TypeLayout<InsnsBuildingDone> {}

        impl<BK: InsnsBuildingKind> TypeLayout<BK> {
            pub(crate) fn len(self) -> TypeLen {
                TypeLen {
                    $($type_field: self.$type_field.len(),)+
                }
            }
            pub(crate) fn is_empty(self) -> bool {
                $(self.$type_field.is_empty())&&+
            }
            pub(crate) fn empty() -> Self {
                Self {
                    $($type_field: StatePartLayout::empty(),)+
                }
            }
        }

        impl TypeLayout<InsnsBuilding> {
            pub(crate) fn next_index(&self) -> TypeIndex {
                TypeIndex {
                    $($type_field: self.$type_field.next_index(),)+
                }
            }
            pub(crate) fn allocate<BK: InsnsBuildingKind>(
                &mut self,
                layout: &TypeLayout<BK>,
            ) -> TypeIndexRange {
                TypeIndexRange {
                    $($type_field: self.$type_field.allocate(&layout.$type_field),)+
                }
            }
        }

        impl From<TypeLayout<InsnsBuilding>> for TypeLayout<InsnsBuildingDone> {
            fn from(v: TypeLayout<InsnsBuilding>) -> Self {
                Self {
                    $($type_field: v.$type_field.into(),)+
                }
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct TypeIndexRange {
            $(pub(crate) $type_field: StatePartIndexRange<$TypeKind>,)+
        }

        impl TypeIndexRange {
            pub(crate) fn new(start: TypeIndex, len: TypeLen) -> Self {
                Self {
                    $($type_field: StatePartIndexRange {
                        start: start.$type_field,
                        len: len.$type_field,
                    },)+
                }
            }
            pub(crate) fn start(self) -> TypeIndex {
                TypeIndex {
                    $($type_field: self.$type_field.start(),)+
                }
            }
            pub(crate) fn len(self) -> TypeLen {
                TypeLen {
                    $($type_field: self.$type_field.len(),)+
                }
            }
            pub(crate) fn is_empty(self) -> bool {
                $(self.$type_field.is_empty()) &&+
            }
            pub(crate) fn slice(self, index: TypeIndexRange) -> Self {
                Self {
                    $($type_field: self.$type_field.slice(index.$type_field),)+
                }
            }
            pub(crate) fn index_array(self, element_size: TypeLen, index: usize) -> Self {
                Self {
                    $($type_field: self.$type_field.index_array(element_size.$type_field, index),)+
                }
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct TypeLen {
            $(pub(crate) $type_field: StatePartLen<$TypeKind>,)+
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        pub(crate) struct TypeIndex {
            $(pub(crate) $type_field: StatePartIndex<$TypeKind>,)+
        }

        #[derive(Debug)]
        pub(crate) struct State {
            pub(crate) insns: Interned<Insns<InsnsBuildingDone>>,
            pub(crate) pc: usize,
            $(pub(crate) $state_field: StatePart<$StateKind>,)+
            $(pub(crate) $type_field: StatePart<$TypeKind>,)+
        }

        impl State {
            pub(crate) fn borrow(&mut self) -> BorrowedState<'_> {
                BorrowedState {
                    orig_insns: self.insns,
                    insns: &self.insns.insns,
                    pc: self.pc,
                    orig_pc: &mut self.pc,
                    $($state_field: self.$state_field.borrow(),)+
                    $($type_field: self.$type_field.borrow(),)+
                }
            }
        }

        #[derive(Debug)]
        pub(crate) struct BorrowedState<'a> {
            pub(crate) orig_insns: Interned<Insns<InsnsBuildingDone>>,
            pub(crate) insns: &'a [Insn],
            pub(crate) orig_pc: &'a mut usize,
            pub(crate) pc: usize,
            $(pub(crate) $state_field: BorrowedStatePart<'a, $StateKind>,)+
            $(pub(crate) $type_field: BorrowedStatePart<'a, $TypeKind>,)+
        }

        impl Drop for BorrowedState<'_> {
            fn drop(&mut self) {
                *self.orig_pc = self.pc;
            }
        }
    };
}

make_state_part_kinds! {
    #[state, field = small_stack]
    impl StatePartKind for StatePartKindSmallStack {
        const NAME: &'static str = "SmallStack";
        type DebugData = ();
        type State = Stack<SmallUInt>;
        type BorrowedState<'a> = BorrowedStack<'a, SmallUInt>;
        fn new_state(len: StatePartLen<Self>) -> Self::State {
            Stack::new(len.value.try_into().expect("state is too big"))
        }
        fn borrow_state<'a>(state: &'a mut Self::State) -> Self::BorrowedState<'a> {
            state.borrow()
        }
    }
    #[state, field = big_stack]
    impl StatePartKind for StatePartKindBigStack {
        const NAME: &'static str = "BigStack";
        type DebugData = ();
        type State = Stack<BigInt>;
        type BorrowedState<'a> = BorrowedStack<'a, BigInt>;
        fn new_state(len: StatePartLen<Self>) -> Self::State {
            Stack::new(len.value.try_into().expect("state is too big"))
        }
        fn borrow_state<'a>(state: &'a mut Self::State) -> Self::BorrowedState<'a> {
            state.borrow()
        }
    }
    #[type, field = small_slots]
    impl StatePartKind for StatePartKindSmallSlots {
        const NAME: &'static str = "SmallSlots";
        type DebugData = SlotDebugData;
        type State = Box<[SmallUInt]>;
        type BorrowedState<'a> = &'a mut [SmallUInt];
        fn new_state(len: StatePartLen<Self>) -> Self::State {
            vec![0; len.value.try_into().expect("state is too big")].into_boxed_slice()
        }
        fn borrow_state<'a>(state: &'a mut Self::State) -> Self::BorrowedState<'a> {
            state
        }
    }
    #[type, field = big_slots]
    impl StatePartKind for StatePartKindBigSlots {
        const NAME: &'static str = "BigSlots";
        type DebugData = SlotDebugData;
        type State = Box<[BigInt]>;
        type BorrowedState<'a> = &'a mut [BigInt];
        fn new_state(len: StatePartLen<Self>) -> Self::State {
            std::iter::repeat_with(BigInt::default).take(len.value.try_into().expect("state is too big")).collect()
        }
        fn borrow_state<'a>(state: &'a mut Self::State) -> Self::BorrowedState<'a> {
            state
        }
    }
}

#[derive(Debug, Clone)]
pub(crate) struct Stack<T> {
    storage: Box<[T]>,
    cur_len: usize,
}

impl<T> Stack<T> {
    pub(crate) fn new(len: usize) -> Self
    where
        T: Default,
    {
        Self {
            storage: std::iter::repeat_with(T::default).take(len).collect(),
            cur_len: 0,
        }
    }
    pub(crate) fn borrow(&mut self) -> BorrowedStack<'_, T> {
        BorrowedStack {
            storage: &mut self.storage,
            cur_len: self.cur_len,
            stack_cur_len: &mut self.cur_len,
        }
    }
    pub(crate) fn clear(&mut self) {
        self.cur_len = 0;
    }
}

#[derive(Debug)]
pub(crate) struct BorrowedStack<'a, T> {
    storage: &'a mut [T],
    cur_len: usize,
    stack_cur_len: &'a mut usize,
}

impl<'a, T> BorrowedStack<'a, T> {
    pub(crate) fn push(&mut self, value: T) {
        self.storage[self.cur_len] = value;
        self.cur_len += 1;
    }
    pub(crate) fn pop(&mut self) -> T
    where
        T: Default,
    {
        self.cur_len -= 1;
        std::mem::take(&mut self.storage[self.cur_len])
    }
}

impl<T> Drop for BorrowedStack<'_, T> {
    fn drop(&mut self) {
        *self.stack_cur_len = self.cur_len;
    }
}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
pub(crate) struct SlotDebugData {
    pub(crate) name: Interned<str>,
}

impl SlotDebugData {
    pub(crate) fn with_prefixed_debug_names(&self, prefix: &str) -> Self {
        let mut name = String::with_capacity(self.name.len() + prefix.len());
        name.push_str(prefix);
        name.push_str(&self.name);
        Self {
            name: Intern::intern_owned(name),
        }
    }
}

impl<K: StatePartKind<DebugData = SlotDebugData>, BK: InsnsBuildingKind> StatePartLayout<K, BK> {
    pub(crate) fn with_prefixed_debug_names(&self, prefix: &str) -> Self {
        Self {
            debug_data: self
                .debug_data
                .iter()
                .map(|v| v.with_prefixed_debug_names(prefix))
                .collect(),
            _phantom: PhantomData,
        }
    }
}

impl<BK: InsnsBuildingKind> TypeLayout<BK> {
    pub(crate) fn with_prefixed_debug_names(&self, prefix: &str) -> Self {
        Self {
            small_slots: self.small_slots.with_prefixed_debug_names(prefix),
            big_slots: self.big_slots.with_prefixed_debug_names(prefix),
        }
    }
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub(crate) struct StatePartIndex<K: StatePartKind> {
    pub(crate) value: u32,
    pub(crate) _phantom: PhantomData<K>,
}

impl<K: StatePartKind> fmt::Debug for StatePartIndex<K> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "StatePartIndex<{}>({})", K::NAME, self.value)
    }
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub(crate) struct StatePartLen<K: StatePartKind> {
    pub(crate) value: u32,
    pub(crate) _phantom: PhantomData<K>,
}

impl<K: StatePartKind> StatePartLen<K> {
    pub(crate) const fn new(value: u32) -> Self {
        Self {
            value,
            _phantom: PhantomData,
        }
    }
}

impl<K: StatePartKind> fmt::Debug for StatePartLen<K> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "StatePartLen<{}>({})", K::NAME, self.value)
    }
}

#[derive(Clone, PartialEq, Eq, Hash)]
pub(crate) struct StatePartLayout<K: StatePartKind, BK: InsnsBuildingKind> {
    pub(crate) debug_data: BK::Vec<K::DebugData>,
    pub(crate) _phantom: PhantomData<K>,
}

impl<K: StatePartKind> Copy for StatePartLayout<K, InsnsBuildingDone> {}

impl<K: StatePartKind, BK: InsnsBuildingKind> StatePartLayout<K, BK> {
    pub(crate) fn len(&self) -> StatePartLen<K> {
        StatePartLen::new(
            self.debug_data
                .len()
                .try_into()
                .expect("state part allocation layout is too big"),
        )
    }
    pub(crate) fn is_empty(&self) -> bool {
        self.debug_data.is_empty()
    }
    pub(crate) fn empty() -> Self {
        Self {
            debug_data: Default::default(),
            _phantom: PhantomData,
        }
    }
}

impl<K: StatePartKind> From<StatePartLayout<K, InsnsBuilding>>
    for StatePartLayout<K, InsnsBuildingDone>
{
    fn from(value: StatePartLayout<K, InsnsBuilding>) -> Self {
        Self {
            debug_data: Intern::intern_owned(value.debug_data),
            _phantom: PhantomData,
        }
    }
}

impl<K: StatePartKind> StatePartLayout<K, InsnsBuilding> {
    pub(crate) fn scalar(debug_data: K::DebugData) -> Self {
        Self {
            debug_data: vec![debug_data],
            _phantom: PhantomData,
        }
    }
    pub(crate) fn next_index(&self) -> StatePartIndex<K> {
        StatePartIndex {
            value: self.len().value,
            _phantom: PhantomData,
        }
    }
    pub(crate) fn allocate<BK: InsnsBuildingKind>(
        &mut self,
        layout: &StatePartLayout<K, BK>,
    ) -> StatePartIndexRange<K> {
        let start = self.next_index();
        let len = layout.len();
        let Self {
            debug_data,
            _phantom: _,
        } = self;
        debug_data.extend_from_slice(&layout.debug_data);
        StatePartIndexRange { start, len }
    }
}

impl<K: StatePartKind, BK: InsnsBuildingKind> fmt::Debug for StatePartLayout<K, BK> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            debug_data,
            _phantom: _,
        } = self;
        write!(f, "StatePartAllocationLayout<{}>", K::NAME)?;
        f.debug_struct("")
            .field("len", &debug_data.len())
            .field("debug_data", debug_data)
            .finish_non_exhaustive()
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash)]
pub(crate) struct StatePartIndexRange<K: StatePartKind> {
    pub(crate) start: StatePartIndex<K>,
    pub(crate) len: StatePartLen<K>,
}

impl<K: StatePartKind> StatePartIndexRange<K> {
    pub(crate) fn start(self) -> StatePartIndex<K> {
        self.start
    }
    pub(crate) fn end(self) -> StatePartIndex<K> {
        StatePartIndex {
            value: self
                .start
                .value
                .checked_add(self.len.value)
                .expect("state part allocation layout is too big"),
            _phantom: PhantomData,
        }
    }
    pub(crate) fn len(self) -> StatePartLen<K> {
        self.len
    }
    pub(crate) fn is_empty(self) -> bool {
        self.len.value == 0
    }
    pub(crate) fn slice(self, index: StatePartIndexRange<K>) -> Self {
        assert!(index.end().value <= self.len.value, "index out of range");
        Self {
            start: StatePartIndex {
                value: self.start.value + index.start.value,
                _phantom: PhantomData,
            },
            len: index.len,
        }
    }
    pub(crate) fn index_array(self, element_size: StatePartLen<K>, index: usize) -> Self {
        if element_size.value == 0 {
            assert_eq!(
                self.len.value, 0,
                "array with zero-sized element must also be zero-sized",
            );
            return self;
        }
        assert!(
            self.len.value % element_size.value == 0,
            "array's size must be a multiple of its element size",
        );
        self.slice(StatePartIndexRange {
            start: StatePartIndex {
                value: index
                    .try_into()
                    .ok()
                    .and_then(|index| element_size.value.checked_mul(index))
                    .expect("index out of range"),
                _phantom: PhantomData,
            },
            len: element_size,
        })
    }
}

impl<K: StatePartKind> fmt::Debug for StatePartIndexRange<K> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "StatePartIndexRange<{}> {{ start: {}, len: {} }}",
            K::NAME,
            self.start.value,
            self.len.value
        )
    }
}

#[derive(Clone, PartialEq, Eq, Hash, Debug)]
pub(crate) struct StatePart<K: StatePartKind> {
    pub(crate) value: K::State,
}

impl<K: StatePartKind> StatePart<K> {
    pub(crate) fn new(len: StatePartLen<K>) -> Self {
        Self {
            value: K::new_state(len),
        }
    }
    pub(crate) fn borrow<'a>(&'a mut self) -> BorrowedStatePart<'a, K> {
        BorrowedStatePart {
            value: K::borrow_state(&mut self.value),
        }
    }
}

#[derive(Clone, PartialEq, Eq, Hash, Debug)]
pub(crate) struct BorrowedStatePart<'a, K: StatePartKind> {
    pub(crate) value: K::BorrowedState<'a>,
}

impl<'a, K: StatePartKind<BorrowedState<'a>: Deref<Target: Index<usize, Output = T>>>, T>
    Index<StatePartIndex<K>> for BorrowedStatePart<'a, K>
{
    type Output = T;
    fn index(&self, index: StatePartIndex<K>) -> &Self::Output {
        &self.value[index.value as usize]
    }
}

impl<'a, K: StatePartKind<BorrowedState<'a>: DerefMut<Target: IndexMut<usize, Output = T>>>, T>
    IndexMut<StatePartIndex<K>> for BorrowedStatePart<'a, K>
{
    fn index_mut(&mut self, index: StatePartIndex<K>) -> &mut Self::Output {
        &mut self.value[index.value as usize]
    }
}

impl<'a, K: StatePartKind<BorrowedState<'a> = BorrowedStack<'a, T>>, T: 'a>
    BorrowedStatePart<'a, K>
{
    pub(crate) fn push(&mut self, value: T) {
        self.value.push(value)
    }
    pub(crate) fn pop(&mut self) -> T
    where
        T: Default,
    {
        self.value.pop()
    }
}

impl Insns<InsnsBuilding> {
    pub(crate) fn new() -> Self {
        Self {
            insns: Vec::new(),
            insn_source_locations: Vec::new(),
            state_layout: StateLayout::empty(),
        }
    }
    pub(crate) fn last_insn_pc(&self) -> usize {
        self.insns
            .len()
            .checked_sub(1)
            .expect("no last instruction")
    }
    pub(crate) fn next_insn_pc(&self) -> usize {
        self.insns.len()
    }
    pub(crate) fn push(&mut self, insn: Insn, source_location: SourceLocation) {
        self.insns.push(insn);
        self.insn_source_locations.push(source_location);
    }
    pub(crate) fn allocate_variable<BK: InsnsBuildingKind>(
        &mut self,
        layout: &TypeLayout<BK>,
    ) -> TypeIndexRange {
        self.state_layout.ty.allocate(layout)
    }
}

impl From<Insns<InsnsBuilding>> for Insns<InsnsBuildingDone> {
    fn from(input: Insns<InsnsBuilding>) -> Self {
        let Insns {
            insns,
            insn_source_locations,
            state_layout,
        } = input;
        Insns {
            insns: Intern::intern_owned(insns),
            insn_source_locations: Intern::intern_owned(insn_source_locations),
            state_layout: state_layout.into(),
        }
    }
}

impl State {
    pub(crate) fn setup_call(&mut self, entry_pc: usize) {
        let Self {
            insns: _,
            pc,
            small_stack,
            big_stack,
            small_slots: _,
            big_slots: _,
        } = self;
        *pc = entry_pc;
        small_stack.value.clear();
        big_stack.value.clear();
    }
}

impl_insns! {
    #[insn = Insn, next_macro = next, branch_macro = branch]
    pub(crate) fn State::run(&mut self) -> () {
        #[state]
        let mut state = self.borrow();
        setup! {}
        main_loop!();
        cleanup! {}
    }
    ReadSmall {
        index: StatePartIndex<StatePartKindSmallSlots>,
    } => {
        state.small_stack.push(state.small_slots[index]);
        next!();
    }
    SExtSmall {
        /// number of bits in a [`SmallSInt`] that aren't used
        unused_bit_count: u8,
    } => {
        let mut value = state.small_stack.pop() as SmallSInt;
        value <<= unused_bit_count;
        value >>= unused_bit_count;
        state.small_stack.push(value as SmallUInt);
        next!();
    }
    ZExtSmall {
        /// number of bits in a [`SmallUInt`] that aren't used
        unused_bit_count: u8,
    } => {
        let mut value = state.small_stack.pop();
        value <<= unused_bit_count;
        value >>= unused_bit_count;
        state.small_stack.push(value);
        next!();
    }
    AndSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = lhs & rhs;
        state.small_stack.push(value);
        next!();
    }
    OrSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = lhs | rhs;
        state.small_stack.push(value);
        next!();
    }
    XorSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = lhs ^ rhs;
        state.small_stack.push(value);
        next!();
    }
    NotSmall => {
        let value = state.small_stack.pop();
        state.small_stack.push(!value);
        next!();
    }
    CmpEqSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = (lhs == rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpNeSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = (lhs != rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpLTSmallUInt => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = (lhs < rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpLESmallUInt => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = (lhs <= rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpGTSmallUInt => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = (lhs > rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpGESmallUInt => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = (lhs >= rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpLTSmallSInt => {
        let rhs = state.small_stack.pop() as SmallSInt;
        let lhs = state.small_stack.pop() as SmallSInt;
        let value = (lhs < rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpLESmallSInt => {
        let rhs = state.small_stack.pop() as SmallSInt;
        let lhs = state.small_stack.pop() as SmallSInt;
        let value = (lhs <= rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpGTSmallSInt => {
        let rhs = state.small_stack.pop() as SmallSInt;
        let lhs = state.small_stack.pop() as SmallSInt;
        let value = (lhs > rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    CmpGESmallSInt => {
        let rhs = state.small_stack.pop() as SmallSInt;
        let lhs = state.small_stack.pop() as SmallSInt;
        let value = (lhs >= rhs) as SmallUInt;
        state.small_stack.push(value);
        next!();
    }
    AddSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = lhs.wrapping_add(rhs);
        state.small_stack.push(value);
        next!();
    }
    SubSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = lhs.wrapping_sub(rhs);
        state.small_stack.push(value);
        next!();
    }
    NegSmall => {
        let value = state.small_stack.pop();
        let value = value.wrapping_neg();
        state.small_stack.push(value);
        next!();
    }
    MulSmall => {
        let rhs = state.small_stack.pop();
        let lhs = state.small_stack.pop();
        let value = lhs.wrapping_mul(rhs);
        state.small_stack.push(value);
        next!();
    }
    ConstU32Small {
        value: u32,
    } => {
        state.small_stack.push(value as SmallUInt);
        next!();
    }
    ConstS32Small {
        value: i32,
    } => {
        state.small_stack.push(value as SmallUInt);
        next!();
    }
    WriteSmall {
        index: StatePartIndex<StatePartKindSmallSlots>,
    } => {
        state.small_slots[index] = state.small_stack.pop();
        next!();
    }
    Return => {
        break;
    }
}
